"""Module to keep methods related to selecting values."""

import random
import uuid

from collections import Iterable

from fauxfactory.helpers import check_validation


def gen_choice(choices):
    """Return a random choice from the available choices.

    :param list choices: List of choices from which select a random value.
    :raises: ``ValueError`` if ``choices`` is ``None`` or not ``Iterable`` or
        a ``dict``.
    :returns: A random element from ``choices``.

    """
    # Validation for 'choices'
    if choices is None:
        raise ValueError('Choices argument cannot be None.')
    # We don't want a single dictionary value.
    if not isinstance(choices, Iterable) or isinstance(choices, dict):
        raise ValueError('Choices argument is not iterable.')
    if not choices:
        raise ValueError('Choices argument cannot be empty.')
    # If only 1 item is present, return it right away
    if len(choices) == 1:
        return choices[0]

    random.seed()
    return random.choice(choices)


@check_validation
def gen_uuid():
    """Generate a UUID string (universally unique identifiers).

    :returns: Returns a string representation for a UUID.
    :rtype: str

    """
    return str(uuid.uuid4())



"""Collection of string generating functions."""
import random
import string

from fauxfactory.constants import HTML_TAGS, LOREM_IPSUM_TEXT
from fauxfactory.helpers import (
    check_len,
    check_validation,
    is_positive_int,
    unicode_letters_generator,
    )


def gen_string(str_type, length=None, validator=None, default=None, tries=10):
    """A simple wrapper that calls other string generation methods.

    :param str str_type: The type of string which should be generated.
    :param int length: The length of the generated string. Must be 1 or
        greater.
    :param validator: Function or regex (str).
        If a function it must receive one parameter and return True if value
        can be used and False of another value need to be generated.
        If str it will be used as regex to validate the generated value.
        Default is None which will not validate the value.
    :param tries: number of times validator must be called before returning
        `default`. Default is 10.
    :param default: If validator returns false a number of `tries` times, this
        value is returned instead. Must be defined if validator is not None
    :raises: ``ValueError`` if an invalid ``str_type`` is specified.
    :returns: A string.
    :rtype: str

    Valid values for ``str_type`` are as follows:

    * alpha
    * alphanumeric
    * cjk
    * cyrillic
    * html
    * latin1
    * numeric
    * utf8
    * punctuation

    """
    str_types_functions = {
        'alpha': gen_alpha,
        'alphanumeric': gen_alphanumeric,
        'cjk': gen_cjk,
        'cyrillic': gen_cyrillic,
        'html': gen_html,
        'latin1': gen_latin1,
        'numeric': gen_numeric_string,
        'utf8': gen_utf8,
        'punctuation': gen_special,
    }
    str_type_lower = str_type.lower()  # do not modify user data
    if str_type_lower not in str_types_functions.keys():
        raise ValueError(
            '{0} is not a supported string type. Valid string types are {1}.'
            ''.format(str_type_lower, ','.join(str_types_functions.keys()))
        )
    method = str_types_functions[str_type_lower]
    if length is None:
        return method(validator=validator, default=default, tries=tries)
    return method(length, validator=validator, default=default, tries=tries)


@check_len
@check_validation
def gen_alpha(length=10):
    """Return a random string made up of alpha characters.

    :param int length: Length for random data.
    :returns: A random string made up of alpha characters.
    :rtype: str

    """
    random.seed()
    output_string = ''.join(
        random.choice(string.ascii_letters) for i in range(length)
    )

    return output_string


@check_len
@check_validation
def gen_alphanumeric(length=10):
    """Return a random string made up of alpha and numeric characters.

    :param int length: Length for random data.
    :returns: A random string made up of alpha and numeric characters.
    :rtype: str

    """
    random.seed()
    output_string = ''.join(
        random.choice(
            string.ascii_letters + string.digits
        ) for i in range(length))

    return output_string


@check_len
@check_validation
def gen_cjk(length=10):
    """Return a random string made up of CJK characters.

    (Source: Wikipedia - CJK Unified Ideographs)

    :param int length: Length for random data.
    :returns: A random string made up of CJK characters.
    :rtype: str

    """
    random.seed()

    # Generate codepoints, then convert the codepoints to a string. The
    # valid range of CJK codepoints is 0x4E00 - 0x9FCC, inclusive.
    codepoints = [random.randint(0x4E00, 0x9FCC) for _ in range(length)]
    return ''.join(chr(codepoint) for codepoint in codepoints)


@check_len
@check_validation
def gen_cyrillic(length=10):
    """Return a random string made up of Cyrillic characters.

    :param int length: Length for random data.
    :returns: A random string made up of Cyrillic characters.
    :rtype: str

    """
    random.seed()

    # Generate codepoints, then convert the codepoints to a string. The
    # valid range of Cyrillic codepoints is 0x410 - 0x4ff, inclusive.
    codepoints = [random.randint(0x0400, 0x04FF) for _ in range(length)]
    return ''.join(chr(codepoint) for codepoint in codepoints)


@check_len
@check_validation
def gen_html(length=10, include_tags=True):
    """Return a random string made up of html characters.

    :param int length: Length for random data.
    :returns: A random string made up of html characters.
    :rtype: str

    """
    random.seed()
    html_tag = random.choice(HTML_TAGS)

    if not include_tags:
        if length < 8:
            raise ValueError(u'Cannot generate html with less than 7 chars')
        maybe_len = length - len('<{0}></{0}>'.format(html_tag))
        if maybe_len <= 0:
            length -= 7
            html_tag = 'a'
        else:
            length = maybe_len
        output_string = u'<{0}>{1}</{2}>'.format(
            html_tag, gen_string('alpha', length), html_tag)
    else:
        output_string = '<{0}>{1}</{2}>'.format(
            html_tag, gen_string('alpha', length), html_tag)

    return output_string


def gen_iplum(words=None, paragraphs=None):
    """Return a lorem ipsum string.

    If no arguments are passed, then return the entire default lorem ipsum
    string.

    :param int words: The number of words to return.
    :param int paragraphs: The number of paragraphs to return.
    :raises: ``ValueError`` if ``words`` is not a valid positive integer.
    :returns: A ``lorem ipsum`` string containing either the number of
        ``words`` or ``paragraphs``, extending and wrapping around the text
        as needed to make sure that it has the specified length.
    :rtype: str

    """
    # Check parameters
    if words is None or words == 0:
        words = len(LOREM_IPSUM_TEXT.split())
    if paragraphs is None:
        paragraphs = 1

    if not isinstance(words, int) or words < 0:
        raise ValueError(
            'Cannot generate a string with negative number of words.')
    is_positive_int(paragraphs)

    # Original Lorem Ipsum string
    all_words = LOREM_IPSUM_TEXT.split()
    # How many words do we need?
    total_words_needed = words * paragraphs

    quotient = int(total_words_needed / len(all_words))
    modulus = total_words_needed % len(all_words)

    # Pool of words to use
    all_words = all_words * (quotient + modulus)

    result = u''
    start_pos = 0
    for _ in range(0, paragraphs):
        sentence = u' '.join(
            all_words[start_pos:start_pos + words])

        # Remove comma from the end, if it exists
        if sentence.endswith(','):
            sentence = sentence.rstrip(',')
        # Remove period from the end, if it exists
        if sentence.endswith('.'):
            sentence = sentence.rstrip('.')

        # Each sentence should be properly capitalized
        cap_sentence = [
            frag.capitalize() + u'.' for frag in sentence.split('. ')]

        # Add newline at the end
        result += ' '.join(cap_sentence) + u'\n'

        # Increment positional counter
        start_pos += words
    return result.rstrip()


@check_len
@check_validation
def gen_latin1(length=10):
    """Return a random string made up of UTF-8 characters.

    (Font: Wikipedia - Latin-1 Supplement Unicode Block)

    :param int length: Length for random data.
    :returns: A random string made up of ``Latin1`` characters.
    :rtype: str

    """
    range0 = range1 = range2 = []
    range0 = ['00C0', '00D6']
    range1 = ['00D8', '00F6']
    range2 = ['00F8', '00FF']
    output_array = []

    for i in range(int(range0[0], 16), int(range0[1], 16)):
        output_array.append(i)
    for i in range(int(range1[0], 16), int(range1[1], 16)):
        output_array.append(i)
    for i in range(int(range2[0], 16), int(range2[1], 16)):
        output_array.append(i)

    random.seed()
    output_string = ''.join(
        chr(random.choice(output_array)) for _ in range(length)
    )

    return output_string


@check_len
@check_validation
def gen_numeric_string(length=10):
    """Return a random string made up of numbers.

    :param int length: Length for random data.
    :returns: A random string made up of numbers.
    :rtype: str

    """
    random.seed()
    return ''.join(random.choice(string.digits) for _ in range(length))


@check_len
@check_validation
def gen_utf8(length=10, smp=True):
    """Return a random string made up of UTF-8 letters characters.

    Follows `RFC 3629`_.

    :param int length: Length for random data.
    :param bool smp: Include Supplementary Multilingual Plane (SMP)
        characters
    :returns: A random string made up of ``UTF-8`` letters characters.
    :rtype: str

    .. _`RFC 3629`: http://www.rfc-editor.org/rfc/rfc3629.txt

    """
    UNICODE_LETTERS = [c for c in unicode_letters_generator(smp)]
    random.seed()
    return ''.join([random.choice(UNICODE_LETTERS) for _ in range(length)])


@check_len
@check_validation
def gen_special(length=10):
    """Return a random special characters string.

    :param int length: Length for random data.
    :returns: A random string made up of special characters.
    :rtype: str
    """
    random.seed()
    return ''.join(random.choice(string.punctuation) for _ in range(length))



"""Collection of computer systems generating functions."""
from copy import deepcopy

from fauxfactory.constants import FACTS_JSON
from fauxfactory.helpers import is_positive_int

from .choices import gen_choice, gen_uuid
from .internet import gen_domain, gen_ipaddr, gen_mac, gen_netmask
from .numbers import gen_integer
from .strings import gen_alpha, gen_alphanumeric


def add_memory_info(count=None):
    """Generate fake memory facts.

    :param int count: The total amount of RAM for a system.
    :returns: A dictionary representing memory facts.
    :rtype: dict

    """
    if count is None:
        count = gen_choice(range(4, 128, 2))
    else:
        is_positive_int(count)

    free_ram = (gen_integer(min_value=4, max_value=count))

    return {
        'dmi::memory::size': '{}'.format(free_ram * 1024),
        'memoryfree': '{} GB'.format(free_ram),
        'memoryfree_mb': '{}'.format(free_ram * 1024),
        'memorysize': '{} GB'.format(count),
        'memorysize_mb': '{}'.format(count * 1024),
        'swapfree': '{} GB'.format(free_ram),
        'swapfree_mb': '{}'.format(free_ram * 1024),
        'swapsize': '{} GB'.format(count),
        'swapsize_mb': '{}'.format(count * 1024),
    }


def add_network_devices():
    """Generate fake network device facts.

    :returns: A dictionary representing a Host's network devices.
    :rtype: dict

    """
    return {
        'dhcp_servers': {
            'system': gen_ipaddr(),
            'enp11s0': gen_ipaddr(),
        },
        'interfaces': 'enp11s0',
        'ipaddress': gen_ipaddr(),
        'ipaddress6': gen_ipaddr(ipv6=True),
        'ipaddress6_enp11s0': gen_ipaddr(ipv6=True),
        'macaddress': gen_mac(),
        'macaddress_enp11s0': gen_mac(),
        'mtu_enp11s0': '1500',
        'netmask': gen_netmask(),
        'netmask_enp11s0': gen_netmask(),
        'netmask_lo': gen_netmask(),
        'network_enp11s0': gen_ipaddr(),
        'network_lo': '127.0.0.0',
    }


def add_operating_system(name=None, family=None, major=None, minor=None):
    """Generate fake operating system facts.

    :param str name: The name for an operating system.
    :param str family: The operating system family.
    :param int major: The major release of the operating system.
    :param int minor: The minor release of the operating system.
    :returns: A dictionary representing an Operating System.
    :rtype: dict

    """
    if name is None:
        name = gen_alpha()
    if family is None:
        family = gen_alpha()
    if major is None:
        major = gen_integer(min_value=0, max_value=9)
    if minor is None:
        minor = gen_integer(min_value=0, max_value=9)
    return {
        'os': {
            'name': name,
            'family': family,
            'release': {
                'major': major,
                'minor': minor,
                'full': '{}.{}'.format(major, minor)
            }
        },
        'operatingsystem': name,
        'operatingsystemmajrelease': major,
        'operatingsystemrelease': '{}.{}'.format(major, minor),
        'osfamily': family,
        'distribution::id': gen_alpha(),
        'distribution::name': name,
        'distribution::version': '{}.{}'.format(major, minor),

    }


def add_partitions(extra_partitions=None):
    """Generate fake partitions facts."""
    partitions = {
        'partitions': {
            'sda1': {
                'uuid': gen_uuid(),
                'size': '1024000',
                'mount': '/boot',
                'filesystem': 'xfs'
            },
        }
    }

    if extra_partitions is not None:
        is_positive_int(extra_partitions)
        for idx in range(extra_partitions):
            device_id = idx + 1
            partitions['partitions'].update(
                {
                    'sdb{}'.format(device_id): {
                        'size': '975747072',
                        'filesystem': 'LVM2_member',
                    }})
    return partitions


def add_processor_info(count=None):
    """Generate fake processor facts.

    :param int count: Number of processors for a system.
    :returns: A dictionary containing fake processor facts.
    :rtype: dict

    """
    if count is None:
        count = gen_choice(range(2, 16, 2))
    else:
        is_positive_int(count)

    processors = {
        'physicalprocessorcount': count,
        'processorcount': count,
        'cpu::topology_source': 'kernel /sys cpu sibling lists',
        'cpu::cpu(s)': count,
        'lscpu::cpu(s)': count,
        'processors': {
            'models': [],
            'count': count,
            'physicalcount': count,
        },
    }

    # Add processors info based on total processors
    for idx in range(count):
        processors['processor{}'.format(
            idx)] = 'Intel(R) Xeon(R) CPU E31220 @ 3.10GHz'
        processors['processors']['models'].append(
            'Intel(R) Xeon(R) CPU E31220 @ 3.10GHz')
    return processors


def gen_system_facts(name=None):
    """Generate system facts.

    See https://docs.puppet.com/facter/3.6/core_facts.html for more
    information.

    :param str name: Name to be used as the system's hostname.
    :returns: A Dictionary representing a system's facts.
    :rtype: dict
    """
    if name is None or name == '':
        fqdn = gen_domain()
    else:
        fqdn = gen_domain(*name.split('.'))

    kernel = '{}.{}.{}'.format(
        *[gen_integer(min_value=0, max_value=9) for _ in range(3)])

    host = deepcopy(FACTS_JSON)

    host['architecture'] = gen_choice(('i386', 'x86_64'))
    host['domain'] = '.'.join(fqdn.split('.')[1:])
    host['fqdn'] = fqdn
    host['hostname'] = fqdn.split('.')[0]

    host['hardwareisa'] = host['architecture']
    host['hardwaremodel'] = host['architecture']
    host['kernelmajversion'] = '.'.join(kernel.split('.')[:2])
    host['kernelrelease'] = '{}-{}.{}'.format(
        kernel,
        gen_integer(min_value=0, max_value=999),
        host['architecture']
    )
    host['kernelversion'] = kernel

    host.update(add_memory_info())

    host.update(add_network_devices())

    host.update(add_operating_system())

    host.update(add_partitions())

    host.update(add_processor_info())

    host['path'] = (
        '/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/root/bin:/sbin')

    host['productname'] = '{}'.format(gen_alpha())
    host['dmi::system::product_name'] = host['productname']
    host['dmi::baseboard::product_name'] = host['productname']
    host['serialnumber'] = '{}'.format(gen_alphanumeric())
    host['timezone'] = 'EDT'
    host['uniqueid'] = '{}'.format(gen_alphanumeric())
    host['uptime_days'] = gen_integer(min_value=1, max_value=1974)
    host['uptime_hours'] = host['uptime_days'] * 24
    host['uptime_seconds'] = host['uptime_hours'] * 3600
    host['uptime'] = '{} days'.format(host['uptime_days'])
    host['system_uptime'] = {
        'seconds': host['uptime_seconds'],
        'hours': host['uptime_hours'],
        'days': host['uptime_days'],
        'uptime': host['uptime']}
    host['uuid'] = gen_uuid()
    host['dmi::system::uuid'] = host['uuid']

    return host



"""Methods that generate random number values."""

from functools import partial
import random
import sys

from fauxfactory.helpers import base_repr


def gen_integer(min_value=None, max_value=None):
    """Return a random integer value based on the current platform.

    :param int min_value: The minimum allowed value.
    :param int max_value: The maximum allowed value.
    :raises: ``ValueError`` if arguments are not integers or if they are
        less or greater than the system's allowed range for integers.
    :returns: Returns a random integer value.
    :rtype: int

    """
    # Platform-specific value range for integers
    _min_value = - sys.maxsize - 1
    _max_value = sys.maxsize

    if min_value is None:
        min_value = _min_value
    if max_value is None:
        max_value = _max_value

    integer_types = (int,)

    # Perform some validations
    if not isinstance(min_value, integer_types) or min_value < _min_value:
        raise ValueError('\'{0}\' is not a valid minimum.'.format(min_value))
    if not isinstance(max_value, integer_types) or max_value > _max_value:
        raise ValueError('\'{0}\' is not a valid maximum.'.format(max_value))

    random.seed()
    value = random.randint(min_value, max_value)

    return value


def gen_negative_integer():
    """Return a random negative integer based on the current platform.

    :returns: Returns a random negative integer value.
    :rtype: int

    """
    max_value = 0

    return gen_integer(max_value=max_value)


def gen_positive_integer():
    """Return a random positive integer based on the current platform.

    :returns: A random positive integer value.
    :rtype: int

    """
    min_value = 0

    return gen_integer(min_value=min_value)


def gen_number(min_value=None, max_value=None, base=10):
    """Return a random number (with <base> representation).

    :returns: A random number with base of <base>.
    :rtype: str

    """
    return base_repr(gen_integer(
        int(str(min_value), base=base) if min_value is not None else min_value,
        int(str(max_value), base=base) if max_value is not None else max_value,
    ), base)


gen_octagonal = partial(gen_number, base=8)
gen_hexadecimal = partial(gen_number, base=16)



"""Collection of helper methods and functions."""
import re
import unicodedata

from collections import namedtuple
from functools import wraps
from fauxfactory.constants import VALID_DIGITS


UnicodePlane = namedtuple('UnicodePlane', ['min', 'max'])

BMP = UnicodePlane(int('0x0000', 16), int('0xffff', 16))
SMP = UnicodePlane(int('0x10000', 16), int('0x1ffff', 16))


def base_repr(number, base):
    """Return the base representation of a decimal number.

    As shared here: https://stackoverflow.com/a/2267446

    Conversion steps:

    1) Divide the number by the base
    2) Get the integer quotient for the next iteration
    3) Get the remainder for the hex digit
    4) Repeat the steps until quotient is equal to zero

    :param number: (int) The decimal number to be converted.
    :param base: The base to convert.

    :return: The base representation of <number>.
    """
    if base <= 1:
        raise ValueError('Cannot represent number with base smaller than 2.')
    if number < 0:
        sign = -1
    elif number == 0:
        return VALID_DIGITS[0]
    else:
        sign = 1

    number *= sign
    digits = []

    while number:
        digits.append(VALID_DIGITS[number % base])
        number //= base

    if sign < 0:
        digits.append('-')

    digits.reverse()

    return ''.join(digits)


def check_len(fnc):
    """Validate generators requiring a `length` argument."""
    @wraps(fnc)
    def wrapped(*args, **kwargs):
        """"Make sure that we verify the `length` argument."""
        if args and len(args) == 1:
            is_positive_int(args[0])
        if 'length' in kwargs.keys():
            if kwargs.get('length') is not None:
                is_positive_int(kwargs.get('length'))

        result = fnc(*args, **kwargs)

        return result
    return wrapped


def check_validation(fcn):
    """Decorate functions requiring validation.

    Simple decorator to validate values generated by function `fnc`
    according to parameters `validator`, `default` and `tries`.

    :param fcn: function to be enhanced
    :return: decorated function
    """
    @wraps(fcn)
    def validate(*args, **kwargs):
        """Perform the validation on decorated function."""
        validator = kwargs.get('validator')
        default = kwargs.get('default')
        tries = kwargs.get('tries', 10)
        if validator and default is None:
            raise ValueError('If "validator" param is defined, "default" '
                             'parameter must not be None')
        if validator is None:
            def validator_fcn(_):
                """No validation passed."""
                return True
        else:
            validator_fcn = validator

        if not callable(validator_fcn):
            def regex_validator(value):
                """Perform RegEx validation."""
                return re.match(validator, value)

            validator_fcn = regex_validator

        # Removing params related to validation but not fcn
        for key in ('validator', 'default', 'tries'):
            if key in kwargs:
                kwargs.pop(key)

        for _ in range(tries):
            value = fcn(*args, **kwargs)
            if validator_fcn(value):
                return value

        return default

    return validate


def is_positive_int(length):
    """Check that `length` argument is an integer greater than zero.

    :param int length: The desired length of the string
    :raises: `ValueError` if `length` is not an `int` or is less than 1.

    """
    if not isinstance(length, int):
        raise ValueError('{0} is not numeric.'.format(length))
    if length <= 0:
        raise ValueError('{0} is an invalid \'length\'.'.format(length))


def unicode_letters_generator(smp=True):
    """Generate unicode characters in the letters category.

    :param bool smp: Include Supplementary Multilingual Plane (SMP)
        characters
    :return: a generator which will generates all unicode letters available

    """
    for i in range(BMP.min, SMP.max if smp else BMP.max):
        char = chr(i)
        if unicodedata.category(char).startswith('L'):
            yield char



"""Constants used by :mod:`fauxfactory`.

.. data:: VALID_NETMASKS

    A tuple of netmasks. The tuple index corresponds to a CIDR value. For
    example, a CIDR of "/1" corresponds to `VALID_NETMASKS[1]`.

"""
# The above constant descriptions can be found by Sphinx and via help().
import datetime
import json
import os
import string


VALID_DIGITS = string.digits + string.ascii_letters

FACTS_JSON = json.load(
    open(os.path.join(
        os.path.dirname(__file__),
        'facts.json')))

LOREM_IPSUM_TEXT = (
    'Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do '
    'eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad '
    'minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip '
    'ex ea commodo consequat. Duis aute irure dolor in reprehenderit in '
    'voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur '
    'sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt '
    'mollit anim id est laborum.'
)

MIN_YEARS = datetime.MINYEAR
MAX_YEARS = 1000  # 1000 years into the future

SCHEMES = (
    'http',
    'https',
    'ftp',
)

SUBDOMAINS = (
    'example',
    'test',
)

TLDS = (
    'biz',
    'com',
    'edu',
    'gov',
    'info',
    'org',
)

VALID_NETMASKS = (
    u'0.0.0.0',
    u'128.0.0.0',
    u'192.0.0.0',
    u'224.0.0.0',
    u'240.0.0.0',
    u'248.0.0.0',
    u'252.0.0.0',
    u'254.0.0.0',
    u'255.0.0.0',
    u'255.128.0.0',
    u'255.192.0.0',
    u'255.224.0.0',
    u'255.240.0.0',
    u'255.248.0.0',
    u'255.252.0.0',
    u'255.254.0.0',
    u'255.255.0.0',
    u'255.255.128.0',
    u'255.255.192.0',
    u'255.255.224.0',
    u'255.255.240.0',
    u'255.255.248.0',
    u'255.255.252.0',
    u'255.255.254.0',
    u'255.255.255.0',
    u'255.255.255.128',
    u'255.255.255.192',
    u'255.255.255.224',
    u'255.255.255.240',
    u'255.255.255.248',
    u'255.255.255.252',
    u'255.255.255.254',
    u'255.255.255.255',
)

HTML_TAGS = (
    'a', 'abbr', 'acronym', 'address', 'applet', 'area', 'b',
    'base', 'basefont', 'bdo', 'big', 'blink', 'blockquote', 'body', 'br',
    'button', 'caption', 'center', 'cite', 'code', 'col', 'colgroup',
    'dd', 'del', 'dfn', 'dir', 'div', 'dl', 'dt',
    'em', 'fieldset', 'font', 'form', 'frame', 'frameset', 'h1',
    'h2', 'h3', 'h4', 'h5', 'h6', 'head', 'hr',
    'html', 'i', 'iframe', 'img', 'input', 'ins', 'isindex',
    'kbd', 'label', 'legend', 'li', 'link', 'map', 'menu',
    'meta', 'noframes', 'noscript', 'object', 'ol', 'optgroup', 'option',
    'p', 'param', 'pre', 'q', 's', 'samp', 'script',
    'select', 'small', 'span', 'strike', 'strong', 'style', 'sub',
    'sup', 'table', 'tbody', 'td', 'textarea', 'tfoot', 'th',
    'thead', 'title', 'tr', 'tt', 'u', 'ul', 'var',
)



"""Generate random data for your tests."""
from importlib import import_module
import os

methods = []

# location for factory modules
factory_dir = os.path.join(os.path.dirname(__file__), 'factories')

# Loop through all factory modules...
for d in os.listdir(factory_dir):
    # ... check that we have a '.py' file...
    if d.endswith('.py') and not d.startswith('__'):
        # Module name is '.factories.<module_name>' without '.py'
        name = ''.join(['.factories.', d[:-3]])
        # Import the module
        module = import_module(name, __package__)
        # Fetch all of its attributes
        for attr in module.__dict__.keys():
            # If attribute name starts with 'gen_'...
            if attr.startswith('gen_'):
                # Add reference to it to locals()
                locals()[attr] = getattr(module, attr)
                # Keep track of all methods being added
                methods.append(attr)

# Add all method names to __all__
__all__ = methods



"""Method for generating random boolean values."""

from .choices import gen_choice


def gen_boolean():
    """Return a random Boolean value.

    :returns: A random Boolean value.
    :rtype: bool

    """
    choices = (True, False)

    return gen_choice(choices)



"""Methods related to generating internet related values."""

import random
import re

from fauxfactory.constants import SCHEMES, SUBDOMAINS, TLDS, VALID_NETMASKS
from fauxfactory.helpers import check_validation

from .choices import gen_choice
from .strings import gen_alpha


def gen_domain(name=None, subdomain=None, tlds=None):
    """Generate a random domain name.

    :param str name: Name for your host.
    :param str subdomain: Name for the subdomain.
    :param str tlds: Top Level Domain Server.
    :returns: A random domain name.
    :rtype: str

    """
    # Generate a new name if needed
    if name is None:
        name = gen_alpha(8).lower()
    # Obtain a random subdomain if needed
    if subdomain is None:
        subdomain = gen_choice(SUBDOMAINS)
    # Obtain a random top level domain if needed
    if tlds is None:
        tlds = gen_choice(TLDS)

    return '{}.{}.{}'.format(name, subdomain, tlds)


@check_validation
def gen_email(name=None, domain=None, tlds=None):
    """Generate a random email address.

    :param str name: Email name.
    :param str domain: Domain name.
    :param str tlds: Top Level Domain Server.
    :returns: An email address.
    :rtype: str

    """
    # Generate a new name if needed
    if name is None:
        name = gen_alpha(8)
    # Obtain a random domain if needed
    if domain is None:
        domain = gen_choice(SUBDOMAINS)
    # Obtain a random top level domain if needed
    if tlds is None:
        tlds = gen_choice(TLDS)

    email = '{0}@{1}.{2}'.format(name, domain, tlds)

    return email


@check_validation
def gen_ipaddr(ip3=False, ipv6=False, prefix=()):
    """Generate a random IP address.

    You can also specify an IP address prefix if you are interested in
    local network address generation, etc.

    :param bool ip3: Whether to generate a 3 or 4 group IP.
    :param bool ipv6: Whether to generate IPv6 or IPv4
    :param list prefix: A prefix to be used for an IP (e.g. [10, 0, 1]). It
        must be an iterable with strings or integers. Can be left
        unspecified or empty.
    :returns: An IP address.
    :rtype: str
    :raises: ``ValueError`` if ``prefix`` would lead to no random fields at
        all. This means the length that triggers the ``ValueError`` is 4 for
        regular IPv4, 3 for IPv4 with ip3 and 8 for IPv6. It will be raised in
        any case the prefix length reaches or exceeds those values.

    """
    # Set the lengths of the randomly generated sections
    if ipv6:
        rng = 8
    elif ip3:
        rng = 3
    else:
        rng = 4
    prefix = [str(field) for field in prefix]
    # Prefix reduces number of random fields generated, so subtract the length
    # of it from the rng to keep the IP address have correct number of fields
    rng -= len(prefix)
    if rng == 0:
        raise ValueError(
            'Prefix {} would lead to no randomness at all'.format(
                repr(prefix)))
    elif rng < 0:
        raise ValueError(
            'Prefix {} is too long for this configuration'.format(
                repr(prefix)))

    random.seed()

    if ipv6:
        # StackOverflow.com questions: generate-random-ipv6-address
        random_fields = [
            '{0:x}'.format(random.randint(0, 2 ** 16 - 1)) for _ in range(rng)]
        ipaddr = ':'.join(prefix + random_fields)
    else:
        random_fields = [str(random.randrange(0, 255, 1)) for _ in range(rng)]
        ipaddr = '.'.join(prefix + random_fields)

        if ip3:
            ipaddr = ipaddr + '.0'

    return ipaddr


@check_validation
def gen_mac(delimiter=':', multicast=None, locally=None):
    """Generate a random MAC address.

    For more information about how unicast or multicast and globally unique and
    locally administered MAC addresses are generated check this link
    https://en.wikipedia.org/wiki/MAC_address.

    :param str delimeter: Valid MAC delimeter (e.g ':', '-').
    :param bool multicast: Indicates if the generated MAC address should be
        unicast or multicast. If no value is provided a random one will be
        chosen.
    :param bool locally: Indicates if the generated MAC address should be
        globally unique or locally administered. If no value is provided a
        random one will be chosen.
    :returns: A random MAC address.
    :rtype: str

    """
    if delimiter not in [':', '-']:
        raise ValueError('Delimiter is not a valid option: %s' % delimiter)
    random.seed()
    if multicast is None:
        multicast = bool(random.randint(0, 1))
    if locally is None:
        locally = bool(random.randint(0, 1))

    first_octet = random.randint(0, 255)
    if multicast:
        # Ensure that the first least significant bit is 1
        first_octet |= 0b00000001
    else:
        # Ensure that the first least significant bit is 0
        first_octet &= 0b11111110
    if locally:
        # Ensure that the second least significant bit is 1
        first_octet |= 0b00000010
    else:
        # Ensure that the second least significant bit is 0
        first_octet &= 0b11111101

    octets = [first_octet]
    octets.extend(random.randint(0, 255) for _ in range(5))
    mac = delimiter.join(['{0:02x}'.format(octet) for octet in octets])

    return mac


@check_validation
def gen_netmask(min_cidr=1, max_cidr=31):
    """Generate a random valid netmask.

    For more info: http://www.iplocation.net/tools/netmask.php

    :param int min_cidr: Inferior CIDR limit
    :param int max_cidr: Superior CIDR limit
    :returns: The netmask is chosen from
        :data:`fauxfactory.constants.VALID_NETMASKS` respecting the CIDR range
    :rtype: str
    :raises: ``ValueError`` if ``min_cidr`` or ``max_cidr`` have an invalid
        value. For example, ``max_cidr`` cannot be 33.

    """
    if min_cidr < 0:
        raise ValueError(
            'min_cidr must be 0 or greater, but is {0}'.format(min_cidr)
        )
    if max_cidr >= len(VALID_NETMASKS):
        raise ValueError(
            'max_cidr must be less than {0}, but is {1}'
            .format(len(VALID_NETMASKS), max_cidr)
        )
    random.seed()
    return VALID_NETMASKS[random.randint(min_cidr, max_cidr)]


@check_validation
def gen_url(scheme=None, subdomain=None, tlds=None):
    """Generate a random URL address.

    :param str scheme: Either http, https or ftp.
    :param str subdomain: A valid subdmain
    :param str tlds: A qualified top level domain name (e.g. 'com', 'net')
    :raises: ``ValueError`` if arguments are not valid.
    :returns: A random URL address.
    :rtype: str

    """
    # Regex for subdomain names
    subdomainator = re.compile(r'^[a-zA-Z0-9][-\w.~]*$')
    # Regex for URL scheme
    schemenator = re.compile(r'^(https?|ftp)$')
    # Regex for TLDS
    tldsnator = re.compile(r'^[a-zA-Z]{1,3}$')

    if scheme:
        if schemenator.match(scheme) is None:
            raise ValueError('Protocol {0} is not valid.'.format(scheme))
    else:
        scheme = gen_choice(SCHEMES)

    if subdomain:
        if subdomainator.match(subdomain) is None:
            raise ValueError('Subdomain {0} is invalid.'.format(subdomain))
    else:
        subdomain = gen_choice(SUBDOMAINS)

    if tlds:
        if tldsnator.match(tlds) is None:
            raise ValueError('TLDS name {0} is invalid.'.format(tlds))
    else:
        tlds = gen_choice(TLDS)

    url = u'{0}://{1}.{2}'.format(scheme, subdomain, tlds)

    return url



"""Methods related to generating date/time related values."""

import datetime
import random

from fauxfactory.constants import MAX_YEARS, MIN_YEARS


def gen_date(min_date=None, max_date=None):
    """Return a random date value.

    :param min_date: A valid ``datetime.date`` object.
    :param max_date: A valid ``datetime.date`` object.
    :raises: ``ValueError`` if arguments are not valid ``datetime.date``
        objects.
    :returns: Random ``datetime.date`` object.

    """
    _min_value = (datetime.date.today() -
                  datetime.timedelta(365 * MIN_YEARS))
    _max_value = (datetime.date.today() +
                  datetime.timedelta(365 * MAX_YEARS))

    if min_date is None:
        min_date = _min_value
    if max_date is None:
        max_date = _max_value

    # Validation
    if not isinstance(min_date, datetime.date):
        raise ValueError('%s is not a valid datetime.date object')
    if not isinstance(max_date, datetime.date):
        raise ValueError('%s is not a valid datetime.date object')

    # Check that max_date is not before min_date
    assert min_date < max_date

    random.seed()

    # Pick a day between min and max dates
    diff = max_date - min_date
    days = random.randint(0, diff.days)
    date = min_date + datetime.timedelta(days=days)

    return date


def gen_datetime(min_date=None, max_date=None):
    """Return a random datetime value.

    :param min_date: A valid ``datetime.datetime`` object.
    :param max_date: A valid ``datetime.datetime`` object.
    :raises: ``ValueError`` if arguments are not valid ``datetime.datetime``
        objects.
    :returns: Random ``datetime.datetime`` object.

    """
    _min_value = (datetime.datetime.now() -
                  datetime.timedelta(365 * MIN_YEARS))
    _max_value = (datetime.datetime.now() +
                  datetime.timedelta(365 * MAX_YEARS))

    if min_date is None:
        min_date = _min_value
    if max_date is None:
        max_date = _max_value

    # Validation
    if not isinstance(min_date, datetime.datetime):
        raise ValueError('%s is not a valid datetime.datetime object')
    if not isinstance(max_date, datetime.datetime):
        raise ValueError('%s is not a valid datetime.datetime object')

    # Check that max_date is not before min_date
    assert min_date < max_date

    random.seed()

    # Pick a time between min and max dates
    diff = max_date - min_date
    seconds = random.randint(0, diff.days * 3600 * 24 + diff.seconds)

    return min_date + datetime.timedelta(seconds=seconds)


def gen_time():
    """Generate a random time.

    :returns: A random ``datetime.time`` object.

    """
    random.seed()
    return datetime.time(
        random.randint(0, 23),
        random.randint(0, 59),
        random.randint(0, 59),
        random.randint(0, 999999),
    )



"""A setuptools-based script for installing FauxFactory."""
# setuptools is preferred over distutils.
import codecs
import os

from setuptools import find_packages, setup


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    filename = os.path.join(*paths)
    with codecs.open(filename, mode='r', encoding='utf-8') as handle:
        return handle.read()


LONG_DESCRIPTION = (read('README.rst') + '\n\n' +
                    read('AUTHORS.rst') + '\n\n' +
                    read('HISTORY.rst'))

setup(
    name='fauxfactory',
    description='Generates random data for your tests.',
    long_description=LONG_DESCRIPTION,
    version='3.0.1',
    author='Og Maciel',
    author_email='omaciel@ogmaciel.com',
    url='https://github.com/omaciel/fauxfactory',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    keywords='python automation data',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Testing',
    ],
    test_suite='tests',
)
